from __future__ import annotations
import pathlib
import logging
from sqlalchemy.orm import sessionmaker
import xarray as xr
import pandas as pd
from typing import Union, List
from abc import ABC, abstractmethod
from sqlalchemy import create_engine
from crypto_history.utilities.general_utilities import register_factory
from crypto_history.utilities.general_utilities import check_for_write_access

logger = logging.getLogger(__package__)


class AbstractDiskWriteCreator(ABC):
    """Abstract disk-writer creator"""
    @abstractmethod
    def factory_method(self, *args, **kwargs):
        """factory method to create the disk-writer"""
        pass

    def save_coin_history_to_file(self,
                                  dataarray: xr.DataArray,
                                  file_location: Union[
                                      pathlib.Path, str
                                  ],
                                  ohlcv_to_deposit):
        """
        Save coin history to file
        Args:
            dataarray: xr.DataArray, dataarray which needs to be \
             written to the file
            file_location: path to the filw where it should be written
            ohlcv_to_deposit: list of ohlcv_fields to be deposited

        Returns:

        """
        product = self.factory_method(file_location)
        product.write_coin_history_dataset_to_disk(dataarray,
                                                   ohlcv_to_deposit)


@register_factory(section="write_to_disk", identifier="json")
class JSONCreator(AbstractDiskWriteCreator):
    """JSON creator"""
    def factory_method(self, *args, **kwargs) -> ConcreteAbstractDiskWriter:
        return ConcreteJSONWriter(*args, **kwargs)


@register_factory(section="write_to_disk", identifier="sqlite")
class SQLiteCreator(AbstractDiskWriteCreator):
    """SQLite creator"""
    def factory_method(self, *args, **kwargs) -> ConcreteAbstractDiskWriter:
        return ConcreteSQLiteWriter(*args, **kwargs)


class ConcreteAbstractDiskWriter(ABC):
    """Concrete abstract disk writer"""
    @abstractmethod
    def write_coin_history_dataset_to_disk(self, *args, **kwargs):
        pass


class ConcreteJSONWriter(ConcreteAbstractDiskWriter):
    def __init__(self,
                 json_file_destination):
        self.json_file_destination = json_file_destination

    def write_coin_history_dataset_to_disk(self, *args, **kwargs):
        raise NotImplementedError


class ConcreteSQLiteWriter(ConcreteAbstractDiskWriter):
    """Writes the fields to a SQLite database.

    Warnings: It will raise an error if there are multiple types of candles
    """
    def __init__(self,
                 sqlite_db_path,
                 ):
        self.sqlite_db_path = sqlite_db_path
        self.engine = create_engine(
            f'sqlite:///{self.sqlite_db_path}',
            echo=True
        )

    @staticmethod
    def get_df_from_da(dataarray: xr.DataArray,
                       reference_asset: str,
                       ohlcv_field: str) -> pd.DataFrame:
        """
        Gets a df from the dataarray whose axis are \
        col-> coins, rows-> timestamps
        Args:
            dataarray: xr.DataArray which will be converted
            reference_asset: reference asset from the dataarray \
             Note that it can have only 1 reference asset
            ohlcv_field: field for flattening

        Returns:
            pd.DataFrame which contains the 2D
            information of the dataarray

        """
        to_convert_to_pd = dataarray.loc[
                           :, reference_asset, :, ohlcv_field
                           ]
        return to_convert_to_pd.transpose().to_pandas()

    @staticmethod
    def get_sql_table_name(dataarray: xr.DataArray,
                           reference_asset: str,
                           ohlcv_field: str
                           ):
        """
        Gets the name of the SQL table. Name is in the format:
        COIN_HISTORY_{ohlcv}_{reference_asset}_{candle}.
        Args:
            dataarray: xr.DataArray which should be converted
            reference_asset: str, reference asset
            ohlcv_field: str, ohlcv_field

        Returns:
            str, a string with the table of the SQL table

        """
        weights = dataarray.loc[
                  :, reference_asset, :, "weight"
                 ].values.flatten()
        unique_values = \
            set(filter(lambda x: isinstance(x, str), weights))
        assert len(unique_values) == 1, \
            f"More than 1 type of weights found. {unique_values}"
        return f"COIN_HISTORY_{ohlcv_field}_" \
               f"{reference_asset}_{unique_values.pop()}"

    def yield_db_name_from_dataset(self,
                                   dataarray: xr.DataArray,
                                   ohlcv_fields: List[str]):
        """
        Yields the df, name of the table from the dataarray
        Args:
            dataarray: xr.DataArray which has to be entered \
             in the SQLite DB
            ohlcv_fields: list of fields to deposit

        Yields: df, name where each df corresponds to one \
            field from the ohlcv fields

        """
        for reference_asset_da in dataarray.reference_assets:
            reference_asset = reference_asset_da.values.tolist()
            for ohlcv_field in ohlcv_fields:
                df = self.get_df_from_da(dataarray,
                                         reference_asset,
                                         ohlcv_field)
                if df.isnull().values.all():
                    logger.warning(f"All the values in the df of {ohlcv_field}"
                                   f" for reference_asset {reference_asset} "
                                   f"are null")
                    continue

                table_name = self.get_sql_table_name(dataarray,
                                                     reference_asset,
                                                     ohlcv_field
                                                     )
                yield df, table_name

    def write_coin_history_dataset_to_disk(self,
                                           dataarray: xr.DataArray,
                                           ohlcv_to_deposit: List[str]):
        """
        Writes the coin history to SQLite DB

        Args:
            dataarray: xr.DataArray which is going to be written in \
            the SQL DB
            ohlcv_to_deposit: types of ohlcv fields to deposit

        """
        if "weight" in ohlcv_to_deposit:
            raise ValueError("weight not expected in ohlcv_to_deposit")

        if check_for_write_access(
                pathlib.Path(self.sqlite_db_path).parent
        ) is True:
            session = sessionmaker(bind=self.engine)()
            for df, name in self.yield_db_name_from_dataset(dataarray,
                                                            ohlcv_to_deposit):
                try:
                    df.to_sql(name, self.engine)
                    session.commit()
                finally:
                    session.close()
        else:
            raise PermissionError(f"Do not have permissions to "
                                  f"write in {self.sqlite_db_path}")


def write_coin_history_to_file(dataarray: xr.DataArray,
                               creator: AbstractDiskWriteCreator,
                               file_location: Union[pathlib.Path, str],
                               *args, **kwargs):
    """
    Writes the coin history to file
    Args:
        dataarray: xr.DataArray, dataset of the coin history
        creator: AbstractDiskWriteCreator, The creator for the \
         disk-writer class
        file_location: location of the file where it should be written
        *args: arguments to be passed to the creator
        **kwargs: keyword-arguments passed to the creator

    """
    if check_for_write_access(pathlib.Path(file_location).parent) is True:
        creator.save_coin_history_to_file(dataarray,
                                          file_location,
                                          *args,
                                          **kwargs)



import logging
import datetime
import xarray as xr
import numpy as np
import pandas as pd
import contextlib
from typing import List, Dict, Union, Tuple, Type
from crypto_history.stock_market.stock_market_factory import StockMarketFactory
from crypto_history.data_container.data_container_pre import \
    PrimitiveDataArrayOperations
from crypto_history.utilities.exceptions import EmptyDataFrameException
from crypto_history.utilities.general_utilities import TypeVarPlaceHolder

logger = logging.getLogger(__name__)


class TimeStampIndexedDataContainer:
    """Responsible for transforming the Primitive DataArray to a DataArray\
     indexed by the timestamp of choice
     with possibility to approximate for easier handling"""

    def __init__(
        self,
        primitive_full_data_container: PrimitiveDataArrayOperations,
        primitive_reference_data_container: PrimitiveDataArrayOperations,
        aggregate_coordinate_by: str,
    ):
        self.aggregate_coordinate_by = aggregate_coordinate_by
        self.primitive_full_data_container = primitive_full_data_container
        self.primitive_reference_data_container = (
            primitive_reference_data_container
        )
        self.ohlcv_coord_name = "ohlcv_fields"

    @classmethod
    @contextlib.asynccontextmanager
    async def create_time_stamp_indexed_data_container(
        cls,
        exchange_factory: StockMarketFactory,
        base_assets: List,
        reference_assets: List,
        reference_ticker: tuple,
        aggregate_coordinate_by: str,
        ohlcv_fields: List,
        weight: str,
        start_time: Union[str, datetime.datetime, int],
        end_time: Union[str, datetime.datetime, int],
    ):
        """
        The factory for creating the time stamp indexed data container
        Args:
            exchange_factory (StockMarketFactory): The exchange factory
            base_assets(List): List of base assets
            reference_assets(List): List of reference assets
            reference_ticker(tuple): ('xxx', 'yyy') where xxx is the \
                base_asset and yyy is the reference asset \
                for indexing the timestamp
            aggregate_coordinate_by(str): The direction in which the \
                coordinates should be aggregated by
            ohlcv_fields(List): list of ohlcv-fields necessary to capture
            weight(str): weight/interval of the kline/candle
            start_time(str/datetime.datetime/int): start time/date of \
                the candles
            end_time(str/datetime.datetime/int): end time/date of \
                the candles

        Yields:
            TimeStampIndexedDataContainer constructed with above details

        """
        async with PrimitiveDataArrayOperations.\
                create_primitive_data_array_operations(
                    exchange_factory,
                    base_assets,
                    reference_assets,
                    ohlcv_fields,
                    weight,
                    start_time,
                    end_time,
                ) as primtive_data_operations:
            reference_base, reference_quote = reference_ticker
            async with PrimitiveDataArrayOperations.\
                    create_primitive_data_array_operations(
                        exchange_factory,
                        [reference_base],
                        [reference_quote],
                        ohlcv_fields,
                        weight,
                        start_time,
                        end_time,
                    ) as primitive_reference_data_container:
                yield cls(
                    primtive_data_operations,
                    primitive_reference_data_container,
                    aggregate_coordinate_by,
                )

    @staticmethod
    async def get_primitive_xr_dataarray(
        data_container_object: PrimitiveDataArrayOperations,
    ) -> xr.DataArray:
        """
        Gets the primitive xr.DataArray from the data container object
        Returns:
            xr.DataArray: the actual xr.DataArray
        """
        logger.debug(
            f"Getting the primitive dataarray for {data_container_object}"
        )
        return await data_container_object.get_populated_primitive_container()

    async def get_primitive_reference_xr_dataarray(self) -> xr.DataArray:
        """
        Gets the reference data container
        Returns:
            xr.DataArray: the reference data container (eg. ETHBTC history)

        """
        logger.debug("Obtain the primitive reference dataarray")
        return await self.get_primitive_xr_dataarray(
            self.primitive_reference_data_container
        )

    async def get_primitive_full_xr_dataarray(self) -> xr.DataArray:
        """
        Gets the full data container
        Returns:
            xr.DataArray: the complete data container

        """
        logger.debug("Obtain the primitive complete dataarray")
        return await self.get_primitive_xr_dataarray(
            self.primitive_full_data_container
        )

    @staticmethod
    def get_all_unique_values(
        dataarray: xr.DataArray, coord_name: str, field_in_coord: str
    ) -> List:
        """
        Gets all the unique values in the xr.DataArray in the particular\
         item of the coordinate
        Args:
            dataarray (xr.DataArray): whose data is to be selected
            coord_name (str): name of the coordinate which is the \
            primary selector
            field_in_coord (str): name of the item in the coordinate\
             which is the second level of select

        Returns:
            list of all the items in the selected coordinate, \
            field sorted increasingly

        """
        selected_da = dataarray.sel(
            {coord_name: field_in_coord}
        )
        flattened_np_array = selected_da.values.flatten()
        none_removed = flattened_np_array[flattened_np_array != np.array(None)]
        nan_removed = none_removed[~pd.isnull(none_removed)]
        list_of_all_ts = list(set(nan_removed.tolist()))
        list_of_all_ts.sort()
        logger.debug(
            f"The unique values of {coord_name} of field {field_in_coord}"
            f" in the dataarray are {list_of_all_ts}"
        )
        return list_of_all_ts

    async def get_timestamps_of_reference_dataarray(
        self, field_in_coord: str
    ) -> List:
        """
        Obtains the time stamp list from the reference dataarray
        Args:
            field_in_coord (str): the item in the ohlcv_fields which\
             is of interest

        Returns:
            list of the time stamps from the reference dataarray
        """
        reference_dataarray = await self.get_primitive_reference_xr_dataarray()
        logger.debug("Getting the timestamp from the reference dataarray")
        return (
            reference_dataarray.sel(ohlcv_fields=field_in_coord)
            .values.flatten()
            .tolist()
        )

    async def get_timestamps_for_new_dataarray(
        self, do_approximation: bool, dataarray: xr.DataArray,
    ) -> List:
        """Obtains the time stamp for the new dataarray.
        It may either select it from the reference dataarray or\
         get it from the actual dataarray if approximation is not desired
        Args:
            do_approximation (bool): if True, timestamps are approximated and\
             taken from the reference dataarray. if False, timestamps are not\
              approximated. All timestamps from all various tickers will be\
               the coordinate for the timestamp
            dataarray (xr.DataArray): the original dataarray which contains\
             all the timestamps in it

        Returns:
             list of the timestamps for the new dataarray
        """
        if do_approximation is True:
            logger.info(
                "The timestamps are going to be approximated to the reference "
                "dataarray's timestamps"
            )
            return await self.get_timestamps_of_reference_dataarray(
                self.aggregate_coordinate_by
            )
        else:
            logger.info("The timestamps will not be approximated")
            return self.get_all_unique_values(
                dataarray, self.ohlcv_coord_name, self.aggregate_coordinate_by
            )

    async def get_xr_dataarray_indexed_by_timestamps(
        self, do_approximation: bool = True, tolerance_ratio: float = 0.001
    ) -> xr.DataArray:
        """
        Gets the xr.DataSet of the coin histories of the particular chunk.
        Obtains the data and then transforms it to the xr.DataSet
        Args:
            do_approximation (bool): check if the timestamps can be \
            approximated to the reference datarray
            tolerance_ratio (float): the ratio of the maximum tolerance for
            approximations of timestamps
        Returns:
            xr.DataSet data of the coin history
        """
        populated_dataarray = await self.get_primitive_full_xr_dataarray()
        new_df = await self.generate_empty_df_with_new_timestamps(
            do_approximation, populated_dataarray
        )

        tolerance = await self.get_value_of_tolerance(
            do_approximation=do_approximation, tolerance_ratio=tolerance_ratio
        )

        index_of_integrating_ts = self.get_index_of_field(
            populated_dataarray,
            self.ohlcv_coord_name,
            self.aggregate_coordinate_by,
        )

        return self._populate_ts_indexed_dataarray(
            populated_dataarray, new_df, index_of_integrating_ts, tolerance
        )

    async def generate_empty_df_with_new_timestamps(
        self, do_approximation: bool, old_dataarray: xr.DataArray
    ) -> xr.DataArray:
        """
        Generates an empty dataframe with new timestamps in the coordinates
        Args:
            do_approximation (bool): Check if the timestamps can be \
             approximated
            old_dataarray (xr.DataArray): The old dataarray used for reference
            for constructing new dataarray

        Returns:
            xr.DataArray The new empty dataarray which has the coordinates set

        """
        reference_ts = await self.get_timestamps_for_new_dataarray(
            do_approximation, old_dataarray
        )

        new_coordinates = self.get_coords_for_timestamp_indexed_datarray(
            old_dataarray, reference_ts
        )
        logger.debug(
            f"The new dataframe will be initialized with"
            f" coordinates: {new_coordinates}"
        )
        return self.generate_empty_dataarray_indexed_by_timestamp(
            coords=new_coordinates
        )

    async def get_value_of_tolerance(
        self, do_approximation: bool, tolerance_ratio: float = 0.001
    ):
        """
        Gets the value of tolerance used for reindexing
        Args:
            do_approximation (bool): Check if approximation can be made
            tolerance_ratio (float): The ratio of maximum tolerance of time \
            difference for generating \
            coordinates of timestamp

        Returns:
            float, value of the tolerance
        """
        if do_approximation is True:
            return await self.calculate_tolerance_value_from_ratio(
                tolerance_ratio
            )
        else:
            return 0

    def get_coords_for_timestamp_indexed_datarray(
        self,
        old_dataarray: xr.DataArray,
        reference_ts: List[int]
    ) -> Dict:
        """
        Gets the coordinates for the timestamp index dataarray
        Args:
            old_dataarray (xr.DataArray): The old dataarray taken as the\
             building block for the new dataarray
            reference_ts (list): The new timestamps which serve as the\
             references instead of index_number

        Returns:
            dictionary of coordinates for the new dataarray

        """
        index_of_integrating_ts = self.get_index_of_field(
            old_dataarray, self.ohlcv_coord_name, self.aggregate_coordinate_by
        )

        new_ohlcv_fields = old_dataarray[self.ohlcv_coord_name].values.tolist()
        new_ohlcv_fields.pop(index_of_integrating_ts)
        coords = {
            "base_assets": old_dataarray.base_assets,
            "reference_assets": old_dataarray.reference_assets,
            "timestamp": reference_ts,
            "ohlcv_fields": new_ohlcv_fields,
        }
        return coords

    @staticmethod
    def get_index_of_field(
        data_array: xr.DataArray, coord_name: str, item_in_coord: str
    ) -> int:
        """
        Gets the index of the particular field in the coordinate
        Args:
            data_array (xr.DataArray): in which the index is to be identified
            coord_name (str): name of the coordinate
            item_in_coord (str): name of the field in the coordinate

        Returns:
            int, the index number which can be used to identify the item

        """
        return data_array[coord_name].values.tolist().index(item_in_coord)

    @staticmethod
    def generate_empty_dataarray_indexed_by_timestamp(
        coords: Dict,
    ) -> xr.DataArray:
        """
        Generates the new dataarray from the coordinates provided
        Args:
            coords(Dict): the coordinates for the new dataarray

        Returns:
            xr.DataArray the new empty dataarray

        """
        return xr.DataArray(None, coords=coords, dims=coords.keys())

    async def calculate_tolerance_value_from_ratio(
        self, ratio: float = 0.001
    ) -> float:
        """
        Calculates the tolerance value based on the tolerance ratio
        Args:
            ratio (float): ratio of the tolerance

        Returns:
            float, value of the tolerance value

        """
        standard_ts = await self.get_timestamps_of_reference_dataarray(
            self.aggregate_coordinate_by
        )
        standard_diff = np.diff(standard_ts).mean()
        logger.debug(
            f"The standard difference between timestamps while calculating"
            f" timestamps are {standard_diff}"
        )
        return standard_diff * ratio

    @staticmethod
    def get_df_to_insert(
        sub_dataarray: xr.DataArray,
        index_of_integrating_ts: int,
        reference_ts: List,
        tolerance: float,
    ) -> pd.DataFrame:
        """
        Calculates the dataframe from the dataarray with one reduced shape size
        Args:
            sub_dataarray (xr.DataArray): the sub dataarray whose dataframe \
            is to be extracted
            index_of_integrating_ts (int): index of the integrating field \
            in the coordinate
            reference_ts (List): list of the indexes according to what the \
            dataframe needs to be adjusted
            tolerance (float): the value of the tolerance which can be \
            considered while matching indices

        Returns:
            pd.DataFrame the reindexed dataframe which contains the data \
            in the required indices

        Raises:
            EmptyDataFrameException if the sub_dataarray contains\
             only na values
        """
        df = pd.DataFrame(sub_dataarray.values.transpose())
        if not df.isna().all().all():
            df = df.dropna()
            df = df.set_index(index_of_integrating_ts)
            df = df.reindex(
                reference_ts, method="nearest", tolerance=tolerance
            )
            df.sort_index()
            return df
        raise EmptyDataFrameException(
            "The data only contains na values. \n"
            "Dataframe cannot be successfully generated"
        )

    def _populate_ts_indexed_dataarray(
        self,
        old_dataarray: xr.DataArray,
        new_dataarray: xr.DataArray,
        index_of_integrating_ts: int,
        tolerance: float,
    ) -> xr.DataArray:
        """
        Populates the new dataarray with the new indexes
        Args:
            old_dataarray (xr.DataArray): The old dataarray which contains the\
            values to build the new dataframe
            new_dataarray: (xr.DataArray): The new empty dataarray whose\
             coordinates have already been set
            index_of_integrating_ts (int): The index of the integrating ts.\
             Essentially, the index of open_ts in ohlcv_fields
            tolerance: the value of the tolerance for reindexing

        Returns:
            xr.DataArray the dataarray whose indices are of the timestamp\
             and the data is filled

        """
        reference_ts = new_dataarray.timestamp.values.tolist()
        # FixMe This is probably not the quickest way to get a new dataarray.
        # Particularly, for non-approximated method should be reindixable?
        for base_coin_iter in old_dataarray:
            for ref_coin_iter in base_coin_iter:
                try:
                    df_to_insert = self.get_df_to_insert(
                        ref_coin_iter,
                        index_of_integrating_ts,
                        reference_ts,
                        tolerance,
                    )
                    new_dataarray.loc[
                        ref_coin_iter.base_assets.values.tolist(),
                        ref_coin_iter.reference_assets.values.tolist(),
                    ] = df_to_insert
                    logger.debug(
                        "The dataframe has been inserted in"
                        " the new dataframe for"
                        f"{ref_coin_iter.base_assets.values.tolist()}"
                        f"{ref_coin_iter.reference_assets.values.tolist()}."
                    )
                except EmptyDataFrameException:
                    logger.debug(
                        f"Empty dataframe for combination of "
                        f"{ref_coin_iter.base_assets.values.tolist()}"
                        f"{ref_coin_iter.reference_assets.values.tolist()}."
                        f" Skipping the df"
                    )
        return new_dataarray


class TimeAggregatedDataContainer:
    """Aggregates various time-ranges into the data container"""
    def __init__(
        self,
        exchange_factory: StockMarketFactory,
        base_assets: List[str],
        reference_assets: List[str],
        ohlcv_fields: List[str],
        time_range_dict: Dict[Tuple[Union[str, datetime.datetime],
                                    Union[str, datetime.datetime]],
                              str],
        reference_ticker: Tuple = ("ETH", "BTC"),
        aggregate_coordinate_by: str = "open_ts",
    ):
        self.exchange_factory = exchange_factory
        self.base_assets = base_assets
        self.reference_assets = reference_assets
        self.ohlcv_fields = ohlcv_fields
        self.time_range_dict = time_range_dict
        self.reference_ticker = reference_ticker
        self.aggregate_coordinate_by = aggregate_coordinate_by
        self.time_interval_splitter = (
            exchange_factory.create_time_interval_chunks()
        )

    @classmethod
    def create_instance(cls: Type[TypeVarPlaceHolder],
                        exchange_factory: StockMarketFactory,
                        base_assets: List[str],
                        reference_assets: List[str],
                        ohlcv_fields: List[str],
                        time_range_dict:
                        Dict[Tuple[Union[datetime.timedelta, str],
                                   Union[datetime.timedelta, str]],
                             str],
                        reference_ticker: Tuple = ("ETH", "BTC"),
                        aggregate_coordinate_by: str = "open_ts",
                        ) -> TypeVarPlaceHolder:
        """
        Creates the time-aggregator from time-deltas which go
         back from the current time
        Args:
            exchange_factory (StockMarketFactory): The exchange factory
            base_assets (List): base-asset coins
            reference_assets (List): reference-asset coins
            ohlcv_fields (List): ohlcv-fields to calculate
            time_range_dict (Dict): dictionary of time-ranges where the\
            keys are tuples of timedeltas
            reference_ticker (Tuple): reference-ticker for reference timestamps
            aggregate_coordinate_by (str): Timestamp indexed by this value

        Returns:
            TimeAggregatedDataContainer

        """
        if all(map(
                lambda x:
                isinstance(x[0], datetime.timedelta) and
                isinstance(x[1], datetime.timedelta),
                time_range_dict.keys())):
            time_now = datetime.datetime.now()
            time_range_temp_dict = {}
            for (time_start, time_end), candle in time_range_dict.items():
                datetime_start = time_now - time_start
                datetime_end = time_now - time_end
                time_range_temp_dict[(datetime_start,
                                      datetime_end)] = candle
                time_range_dict = time_range_temp_dict
        elif all(map(
                lambda x:
                isinstance(x[0], str) and
                isinstance(x[1], str),
                time_range_dict.keys()
        )
        ):
            pass
        else:
            raise TypeError("Invalid time values")
        return cls(
            exchange_factory,
            base_assets=base_assets,
            reference_assets=reference_assets,
            ohlcv_fields=ohlcv_fields,
            time_range_dict=time_range_dict,
            reference_ticker=reference_ticker,
            aggregate_coordinate_by=aggregate_coordinate_by
        )

    def get_time_interval_chunks(self, time_range: Dict) -> List[tuple]:
        """
        Gets the time interval chunks from the raw dict provided
        Args:
            time_range: Dict with key=tuple of start and end time\
                                  value=type of klines/intervals

        Returns:
            List of tuples with the individual requests for histories
                     ((start time, end time), interval)

        """
        return self.time_interval_splitter.get_time_range_for_historical_calls(
            time_range
        )

    @staticmethod
    def concatenate_dataarray_in_coord(
        *dataarray: Union[List[xr.DataArray], Tuple[xr.DataArray]],
        coordinate="timestamp",
    ) -> xr.DataArray:
        """
        Concatenates two histories in the desired coordinate/dimension
        Args:
            dataarray List[xr.DataArray]: List/Tuple of xr.DataArrays \
                that need to be concatednated
            coordinate: coordinate direction to concat by

        Returns:
            Concatenated xr.DataArray consisting of all the data concatenated

        """
        return xr.concat(
            list(filter(lambda x: x is not None, dataarray)), dim=coordinate
        )

    async def get_chunk_history(
        self,
        kline_width: str,
        start_time: Union[str, datetime.datetime, int],
        end_time: Union[str, datetime.datetime, int],
    ) -> xr.DataArray:
        """
        Gets the history of the particular chunk whose start and end
        times are well defined
        Args:
            kline_width (str): The exchanges interval
            start_time (datetime/str/int): the start-time of the \
                        chunk of history
            end_time: (datetime/str/int): the end-time of the \
                        chunk of history

        Returns:
            xr.DataArray the history of the chunk

        """
        async with TimeStampIndexedDataContainer.\
                create_time_stamp_indexed_data_container(
                    exchange_factory=self.exchange_factory,
                    base_assets=self.base_assets,
                    reference_assets=self.reference_assets,
                    reference_ticker=self.reference_ticker,
                    aggregate_coordinate_by=self.aggregate_coordinate_by,
                    ohlcv_fields=self.ohlcv_fields,
                    weight=kline_width,
                    start_time=start_time,
                    end_time=end_time,
                ) as time_stamp_indexed_container:
            return await time_stamp_indexed_container.\
                get_xr_dataarray_indexed_by_timestamps(
                    do_approximation=False
                )

    async def get_time_aggregated_data_container(
        self, sort: bool = True
    ) -> xr.DataArray:
        """
        Gets time aggregated data container by splitting
         the containers as required

        Args:
            sort (bool): to check if the data is to be sorted \
            in an increasing order

        Returns:
            xr.DataArray The complete history of the interval

        """
        chunks_of_time = self.get_time_interval_chunks(self.time_range_dict)
        chunks = []
        for (start_time, end_time), interval in chunks_of_time:
            chunk_history = await self.get_chunk_history(
                interval, start_time, end_time
            )
            chunks.append(chunk_history)
        overall_history = self.concatenate_dataarray_in_coord(*chunks)
        if sort is True:
            overall_history = self.get_sorted_dataarray(
                overall_history, "timestamp"
            )
        return overall_history

    @staticmethod
    def get_sorted_dataarray(
        dataarray: xr.DataArray,
        coordinate: str
    ) -> xr.DataArray:
        """
        Sorts the given dataarray in the given coordinate direction
        Args:
            dataarray (xr.DataArray): The dataarray which has to be sorted
            coordinate: sort the dataarray in the particular \
                direction of the coordinate

        Returns:
            xr.DataArray The sorted dataarray

        """
        return dataarray.sortby(coordinate, ascending=True)



from __future__ import annotations
import logging
import asyncio
import math
import datetime
import contextlib
from dateutil import parser
from abc import ABC, abstractmethod
from typing import Union, List, Dict, Generator
from functools import lru_cache
from collections import namedtuple
from binance import enums, client
from dataclasses import make_dataclass
from pydoc import locate
from crypto_history.stock_market.tickers import BinanceTickerPool, TickerPool
from crypto_history.stock_market.request import (
    AbstractMarketRequester,
    BinanceRequester,
    SomeOtherExchangeRequester,
)
from crypto_history.utilities.general_utilities import (
    AbstractFactory,
    register_factory,
    get_dataclass_from_dict,
)
from crypto_history.utilities.datetime_operations import \
    DateTimeOperations

logger = logging.getLogger(__name__)


class StockMarketFactory(AbstractFactory):
    """
    Abstract factory to generate factories per exchange
    """

    @staticmethod
    @abstractmethod
    def create_market_requester() -> AbstractMarketRequester:
        """
        Create the low-level market requester instance per exchange
        """
        pass

    @staticmethod
    @abstractmethod
    def create_market_operations() -> AbstractMarketOperations:
        """
        Create the instance of the class to channel the right requests\
         to the low level market requester.

        """
        pass

    @staticmethod
    @abstractmethod
    def create_data_homogenizer() -> AbstractMarketHomogenizer:
        """
        Create an instance of a market homogenizer which is the interface\
         to the market from the outside. It should ensure that the market\
          data from various markets/exchanges which might have different \
          low-level APIs is available in a uniform format.

        """
        pass

    @staticmethod
    @abstractmethod
    def create_ohlcv_field_types() -> AbstractOHLCVFieldTypes:
        pass

    @staticmethod
    @abstractmethod
    def create_time_interval_chunks() -> AbstractTimeIntervalChunks:
        pass


@register_factory(section="market", identifier="binance")
class ConcreteBinanceFactory(StockMarketFactory):
    """Binance's factory for creating factories"""

    @staticmethod
    def create_market_requester() -> BinanceRequester:
        """
        Creates the instance for the Low Level Binance Requester

        Returns:
             BinanceRequester: Instance of BinanceRequester

        """
        return BinanceRequester()

    @contextlib.asynccontextmanager
    async def create_market_operations(self) -> BinanceMarketOperations:
        """
        Creates the instance of the Binance Market Operator

        Returns:
             BinanceMarketOperations: Instance of BinanceMarketOperator

        """
        async with self.create_market_requester() as market_requester:
            yield BinanceMarketOperations(market_requester)

    @contextlib.asynccontextmanager
    async def create_data_homogenizer(self) -> BinanceHomogenizer:
        """
        Creates the instance of the Binance Market Homogenizer

        Returns:
             BinanceHomogenizer: Instance of BinanceHomogenizer

        """
        async with self.create_market_operations() as market_operator:
            type_checker = self.create_ohlcv_field_types()
            yield BinanceHomogenizer(market_operator, type_checker)

    @staticmethod
    def create_ohlcv_field_types() -> BinanceOHLCVFieldTypes:
        """
        Creates the instance of the Binance OHLCV Fields Types

        Returns:
            BinanceOHLCVFieldTypes: Instance of BinanceOHLCVFieldTypes

        """
        return BinanceOHLCVFieldTypes()

    @staticmethod
    def create_time_interval_chunks() -> BinanceTimeIntervalChunks:
        """
        Creates the instance of the Binance time interval chunks

        Returns:
            BinanceTimeIntervalChunks: Instance of BinanceTimeIntervalChunks

        """
        return BinanceTimeIntervalChunks()


@register_factory(section="market", identifier="some_other_exchange")
class ConcreteSomeOtherExchangeFactory(StockMarketFactory):
    """Demo for how another exchange/market's factories would be\
     implemented in this module"""

    @staticmethod
    def create_market_requester() -> SomeOtherExchangeRequester:
        """Creates the instance of another Market Requester"""
        raise NotImplementedError

    @staticmethod
    def create_market_operations() -> SomeOtherExchangeMarketOperations:
        """Creates the instance of another Market's Operator"""
        raise NotImplementedError

    @staticmethod
    def create_data_homogenizer() -> SomeOtherExchangeHomogenizer:
        """Creates the instance of the Market Homogenizer"""
        raise NotImplementedError

    @staticmethod
    def create_ohlcv_field_types() -> SomeOtherOHLCVFieldTypes:
        raise NotImplementedError

    @staticmethod
    def create_time_interval_chunks() -> SomeOtherTimeIntervalChunks:
        raise NotImplementedError


class AbstractMarketOperations(ABC):
    """Abstract Base Class to serve as the parent for all market operators.

    Seeing that the market requester may respond with different formats,\
     it is not possible to know the signature of the method. \
     The arguments and the return values are unknown by the \
     AbstractMarketOperator

    """

    def __init__(self, market_requester):
        """
        The low level market requester which does the API calls

        Args:
            market_requester: instance of the market requester of\
             the corresponding exchange
        """
        self.market_requester = market_requester

    @abstractmethod
    async def get_all_raw_tickers(self, *args, **kwargs):
        """
        Obtain all the tickers from the low-level market requester.
        Seeing that the market requester may respond with different\
         formats, it is not possible to know the signature of the method
        """
        pass

    @abstractmethod
    async def get_raw_history_for_ticker(self, *args, **kwargs):
        """
        Obtain the raw history of a particular ticker.
        """
        pass

    @abstractmethod
    async def get_raw_symbol_info(self, *args, **kwargs):
        """
        Obtains the information on the particular ticker.
        It is not known what are the exact information that is\
         going to be received because the information is coming \
         from various exchanges
        """
        pass

    @abstractmethod
    async def get_exchange_info(self, *args, **kwargs):
        """Gets general properties of the exchange"""
        pass


class BinanceMarketOperations(AbstractMarketOperations):
    """Binance's market operator. Implements methods to get market knowledge"""

    @staticmethod
    def _match_binance_enum(string_to_match: str):
        """
        Convert the string to the python-binance API's enum
        :meth:`binance.enums` in :py:mod:`python-binance`
        https://python-binance.readthedocs.io/en/latest/\
        binance.html#binance.client.AsyncClient

        Args:
            string_to_match(str): string which should be matched\
             to binance's enum

        Returns:
             binance.enum: enum object from python-binance

        """
        binance_matched_enum = list()
        for item in dir(enums):
            if getattr(enums, item) == string_to_match:
                binance_matched_enum.append(item)
        assert (
            len(binance_matched_enum) == 1
        ), f"Multiple Binance enums matched with {string_to_match}"
        return getattr(client.AsyncClient, binance_matched_enum[0])

    async def get_raw_history_for_ticker(
        self,
        ticker: Union[str],
        interval: str,
        start_time: int,
        end_time: int,
    ) -> List:
        """
        Gets the kline history of the ticker from binance exchange

        Args:
            ticker (str): ticker whose history has to be pulled
            interval(str): interval of the history (eg. 1d, 3m, etc).\
            See :meth:`binance.enums` in :py:mod:`python-binance`
            start_time(int): Start date string in exchange-format
            end_time(int): End date string in exchange-format

        Returns:
             list: List of snapshots of history.
             Each snapshot is a list of collection of OHLCV values.
             See details in :class:`.BinanceHomogenizer.OHLCVFields`

        """
        binance_interval = self._match_binance_enum(interval)
        return await self.market_requester.request(
            "get_historical_klines",
            ticker,
            binance_interval,
            start_time,
            end_time,
        )

    async def get_all_raw_tickers(self):
        """
        Gets all the tickers available on the binance exchange.

        Returns:
             list: All the raw tickers obtained from the python-binance\
              (:py:mod:`python-binance`)\
             In binance, the format of each raw ticker is \
             {"symbol": <>, "price": <>}

        """
        return await self.market_requester.request("get_all_tickers")

    async def get_raw_symbol_info(self, symbol: str):
        """
        Obtains the information for the symbol/ticker requested

        Args:
            symbol(str): symbol of the ticker whose information is desired

        Returns:
             dict: The raw information of the ticker desired with \
             information where the keys are the
             baseAsset, precision, quoteAsset, etc.
             See :meth:`binance.AsyncClient.get_symbol_info` in \
             :py:mod:`python-binance
             <https://python-binance.readthedocs.io/en/latest/\
             binance.html#binance.client.Client>`

        """
        return await self.market_requester.request("get_symbol_info", symbol)

    async def get_exchange_info(self):
        """
        Obtains the complete information available at the exchange
        Returns:

        """
        return await self.market_requester.request("get_exchange_info")


class SomeOtherExchangeMarketOperations(AbstractMarketOperations):
    """Place holder for another exchange that could be \
    integrated in this module"""

    def get_raw_symbol_info(self, *args, **kwargs):
        raise NotImplementedError

    async def get_raw_history_for_ticker(self, *args, **kwargs):
        raise NotImplementedError

    async def get_all_raw_tickers(self, *args, **kwargs):
        raise NotImplementedError

    def get_exchange_info(self, *args, **kwargs):
        raise NotImplementedError


class AbstractMarketHomogenizer(ABC):
    """Synthesizes the information obtained from various different market\
     operator to a consistent format"""

    OHLCVFields = None
    # TODO Make it an abstractmethod

    def __init__(self, market_operations, type_checker):
        """
        Initializes the class with the instance of the corresponding \
        market operator

        Args:
            market_operations (AbstractMarketOperations): Instance of \
            the corresponding market operator
        """
        self.market_operator = market_operations
        self.type_checker = type_checker

    @abstractmethod
    async def get_all_coins_ticker_objects(self) -> TickerPool:
        """
        Generates the standard/uniform TickerPool object which should \
        be consistent no matter which exchange it is coming from

        Returns:
             TickerPool: TickerPool which contains all the different \
             tickers and holds it in one

        """
        pass

    @abstractmethod
    def get_ticker_instance(self, *args, **kwargs):
        """
        Gets the standard ticker dataclass object.
        Note that the fields may be dependent on the exchange \
        that it is coming from

        Args:
            *args: unknown, as it depends on the exchange
            **kwargs: unknown, as it depends on the exchange

        Returns:
             dataclass: dataclass instance of the symbol info

        """
        pass

    @abstractmethod
    def get_history_for_ticker(self, *args, **kwargs) -> map:
        """
        Gets the history of the ticker symbol

        Args:
            *args: unknown, as it depends on the exchange
            **kwargs: unknown, as it depends on the exchange

        Returns:
             map: map object of history of the ticker

        """
        pass

    @abstractmethod
    def get_all_base_assets(self):
        pass

    @abstractmethod
    def get_all_reference_assets(self):
        pass

    @abstractmethod
    def get_all_raw_tickers(self):
        pass

    def get_named_ohlcv_tuple(self):
        return self.type_checker.get_named_tuple()


class BinanceHomogenizer(AbstractMarketHomogenizer):
    async def get_exchange_assets(self, type_of_asset: str) -> Generator:
        """
        Obtains the type of asset from the exchange.
        Args:
            type_of_asset (str): string identifier that the binance\
             API uses to identify the key in each symbol

        Returns:
            Generator: The generator of items that are available in\
             the exchange

        """
        exchange_info = await self.get_exchange_info()
        symbols_on_exchange = exchange_info.symbols
        return (symbol[type_of_asset] for symbol in symbols_on_exchange)

    async def get_all_base_assets(self) -> List:
        """Obtains the list of all base assets """
        return list(set(await self.get_exchange_assets("baseAsset")))

    async def get_all_reference_assets(self) -> List:
        """Obtains the list of all reference assets"""
        return list(set(await self.get_exchange_assets("quoteAsset")))

    async def get_all_raw_tickers(self) -> List:
        """Obtains the list of all raw tickers available on binance"""
        return await self.market_operator.get_all_raw_tickers()

    async def get_exchange_info(self):
        exchange_dict = await self.market_operator.get_exchange_info()
        exchange_dataclass = get_dataclass_from_dict(
            "exchange_info", exchange_dict
        )
        return exchange_dataclass

    async def get_set_of_ticker_attributes(self, attribute: str):
        """
        Aggregates the set items of the required attribute
        across the whole history

        Args:
            attribute: The attribute/key which is the common term \
            in the history whose values are to be identified

        Returns:
             set: set of values whose attributes are common

        """

        values = set()
        for ticker in await self.get_all_coins_ticker_objects():
            values.add(getattr(ticker, attribute))
        return values

    @lru_cache(maxsize=1)
    async def get_all_coins_ticker_objects(self) -> TickerPool:
        """
        Obtains the exchange-independent TickerPool
        from all the tickers/symbols that are available on the exchange

        Returns:
             TickerPool: All tickers and their information stored in the\
              TickerPool defined in
        # TODO

        """
        all_raw_tickers = await self.get_all_raw_tickers()
        gathered_operations = []
        for raw_ticker in all_raw_tickers:
            gathered_operations.append(
                self.get_ticker_instance(raw_ticker["symbol"])
            )
        all_tickers = await asyncio.gather(
            *gathered_operations, return_exceptions=False
        )
        return BinanceTickerPool(all_tickers)

    @staticmethod
    def _get_ticker_dataclass(symbol_info_dict: Dict):
        """
        Generates a dataclass from the dictionary of the raw ticker provided

        Args:
            symbol_info_dict: dictionary of the raw ticker

        Returns:
             Ticker: A dataclass instance which holds the symbol information

        """
        return make_dataclass(
            "Ticker",
            fields={k: type(v) for k, v in symbol_info_dict.items()},
            eq=True,
            frozen=True,
        )

    async def get_ticker_instance(self, ticker_name: str):
        """
        Obtains the TickerDataclass based on the string of the \
        ticker name provided

        Args:
            ticker_name(str): ticker name whose standard Ticker \
            dataclass object is desired

        Returns:
             dataclass: instance of the dataclass of the ticker

        """
        symbol_info_dict = await self.market_operator.get_raw_symbol_info(
            ticker_name
        )
        data_class_instance = self._get_ticker_dataclass(symbol_info_dict)
        return data_class_instance(**symbol_info_dict)

    async def get_history_for_ticker(self, *args, **kwargs) -> map:
        """
        Gets the history of the ticker for the desired duration,

        # TODO Probably yank the doc from MarketOperator to here

        Args:
            *args: See the :class:`.BinanceMarketOperator.\
            get_raw_history_for_ticker` for arguments
            **kwargs: See the :class:`.BinanceMarketOperator.\
            get_raw_history_for_ticker` for arguments

        Returns:
             map: map of the history of the ticker mapped to \
             the OHLCVFields namedtuple

        """
        raw_history = await self.market_operator.get_raw_history_for_ticker(
            *args, **kwargs
        )
        named_tuple_instance = self.get_named_ohlcv_tuple()
        return map(lambda x: named_tuple_instance(*x), raw_history)


class SomeOtherExchangeHomogenizer(AbstractMarketHomogenizer):
    """Placeholder to show how another market\
     homogenizer could be implemented"""

    def get_ticker_instance(self, *args, **kwargs):
        raise NotImplementedError

    def get_all_coins_ticker_objects(self) -> TickerPool:
        raise NotImplementedError

    def get_history_for_ticker(self, *args, **kwargs):
        raise NotImplementedError

    def get_all_reference_assets(self):
        raise NotImplementedError

    def get_all_base_assets(self):
        raise NotImplementedError

    def get_all_raw_tickers(self):
        raise NotImplementedError


class AbstractOHLCVFieldTypes(ABC):
    class OHLCVFields:
        pass

    # Do not consider a dataclass as I could not
    # find a quick way to get data in a df from data
    def get_named_tuple(self):
        return namedtuple(
            self.OHLCVFields.__name__, self.OHLCVFields.__annotations__.keys()
        )

    def get_dict_name_type(self):
        return dict(
            map(
                lambda x: (x[0], locate(x[1])),
                self.OHLCVFields.__annotations__.items(),
            )
        )


class BinanceOHLCVFieldTypes(AbstractOHLCVFieldTypes):
    """Fields for the named tuple of the OHLCV returned by Binance \
    get_klines_history
            See Also: https://github.com/binance-exchange/\
            binance-official-api-docs/blob/master/\
            rest-api.md#klinecandlestick-data
            See :meth:`binance.AsyncClient.get_historical_klines` in \
            :py:mod:`python-binance`
    """

    class OHLCVFields:
        open_ts: int
        open: float
        high: float
        low: float
        close: float
        volume: float
        close_ts: int
        quote_asset_value: float
        number_of_trades: int
        taker_buy_base_asset_value: float
        take_buy_quote_asset_value: float
        ignored: int


class SomeOtherOHLCVFieldTypes(AbstractOHLCVFieldTypes):
    pass


class AbstractTimeIntervalChunks(ABC):
    """Abstract class to handle the chunking of dates as the \
    exchanges have limits on the length of the interval of history"""

    limit = math.inf
    url = ""

    def get_time_range_for_historical_calls(
        self, raw_time_range_dict: Dict
    ) -> List[tuple]:
        """
        Obtains the time ranges for making the historical calls.

        Args:
            raw_time_range_dict: Dictionary of the
            (start:str, end:str): (type_of_interval: str)

        Returns:
            List of the tuples of the exchange-specific format.
            ((start_time, end_time), type_of_interval)

        """
        final_time_range = []
        datetime_operations = DateTimeOperations()
        for (
            (start_time, end_time),
            type_of_interval,
        ) in raw_time_range_dict.items():
            if isinstance(start_time, str):
                start_time = self.sanitize_item_to_datetime_object(start_time)
            if isinstance(end_time, str):
                end_time = self.sanitize_item_to_datetime_object(end_time)
            sanitized_kline_width = datetime_operations.\
                map_string_to_timedelta(
                    type_of_interval
                )
            sub_chunks = self._get_chunks_from_start_end_complete(
                start_time, end_time, sanitized_kline_width
            )
            sanitized_sub_chunks = self.get_exchange_specific_sub_chunks(
                sub_chunks
            )
            [
                final_time_range.append((chunk, type_of_interval))
                for chunk in sanitized_sub_chunks
            ]
        logger.info(
            f"The time histories have been chunked into" f" {final_time_range}"
        )
        return final_time_range

    def get_exchange_specific_sub_chunks(
        self, sub_chunks: List[tuple]
    ) -> List[tuple]:
        """
        Gets the exchange specific sub-chunk from default sub-chunks
        Args:
            sub_chunks (list): default sub-chunks to be converted \
            to exchange specific

        Returns:
            List of exchange-specific formatted sub-chunks

        """
        list_of_sub_chunks = []
        for sub_chunk in sub_chunks:
            list_of_sub_chunks.append(
                tuple(
                    map(self.sanitize_datetime_to_exchange_specific, sub_chunk)
                )
            )
        return list_of_sub_chunks

    @staticmethod
    def sanitize_item_to_datetime_object(
        item_to_parse: str,
    ) -> datetime.datetime:
        """
        Converts/sanitizes the string to the datetime.datetime object
        Notes:
            Timezone is not handled. The local timezone is considered \
            by dateutil's parser
        Args:
            item_to_parse (str): the item that has to be converted

        Returns:
            datetime.datetime object from the string

        """
        if item_to_parse == "now":
            return datetime.datetime.now()
        return parser.parse(item_to_parse)

    @staticmethod
    def _calculate_klines_in_interval(
        start_datetime: datetime.datetime,
        end_datetime: datetime.datetime,
        timedelta_kline: datetime.timedelta,
    ) -> int:
        """
        Calculates the expected klines from the start and end dates
        Args:
            start_datetime (datetime.datetime): start time
            end_datetime (datetime.datetime): end time
            timedelta_kline (datetime.timedelta): the interval of each kline

        Returns:
            The number of klines expected between these times

        """
        length_of_interval = end_datetime - start_datetime
        return math.ceil(length_of_interval / timedelta_kline)

    @staticmethod
    def _get_chunks_from_number_of_intervals(
        start_datetime: datetime.datetime,
        end_datetime: datetime.datetime,
        number_of_intervals: int,
    ) -> List[tuple]:
        """
        Gets the chunks by splitting the times into the number of intervals
        Args:
            start_datetime (datetime.datetime): start of the overall time
            end_datetime (datetime.datetime): end of the overall time
            number_of_intervals (int): the number of intervals between the \
             times

        Returns:
            List[tuple] the tuple is set as (start_ts_of_sub_chunk, \
            end_ts_of_sub_chunk)

        """
        chunks_of_intervals = []
        average_timedelta = (
            end_datetime - start_datetime
        ) / number_of_intervals
        for i in range(number_of_intervals):
            chunks_of_intervals.append(
                (
                    start_datetime + average_timedelta * i,
                    start_datetime + average_timedelta * (i + 1),
                )
            )
        # There could possibly be a difference in the end-time of the
        # last time item due to round-off errors
        return chunks_of_intervals

    def _get_chunks_from_start_end_complete(
        self,
        start_datetime: datetime.datetime,
        end_datetime: datetime.datetime,
        interval_kline: datetime.timedelta,
    ) -> List[tuple]:
        """
        Gets the chunks based on the intervals of the timedelta
        Args:
            start_datetime (datetime.datetime): start of the overall time
            end_datetime (datetime.datetime): end of the overall time
            interval_kline (datetime.timedelta): the timedelta of each kline

        Returns:
            List[tuple] the tuple is set as (start_ts_of_sub_chunk, \
            end_ts_of_sub_chunk)

        """
        number_of_klines = self._calculate_klines_in_interval(
            start_datetime, end_datetime, interval_kline
        )
        number_of_intervals_necessary = math.ceil(
            number_of_klines / self.limit
        )

        return self._get_chunks_from_number_of_intervals(
            start_datetime, end_datetime, number_of_intervals_necessary
        )

    @staticmethod
    @abstractmethod
    def sanitize_datetime_to_exchange_specific(
        datetime_obj: datetime.datetime,
    ):
        """Converts the datetime object to exchange specific format"""
        pass


class BinanceTimeIntervalChunks(AbstractTimeIntervalChunks):
    """Binance specific information for the interval generation"""

    url = "https://github.com/binance-exchange/binance-\
    official-api-docs/blob/master/rest-api.md#enum-definitions"
    limit = 1000

    @staticmethod
    def sanitize_datetime_to_exchange_specific(
        datetime_obj: datetime.datetime,
    ) -> int:
        """
        Converts the datetime object to binance specific format for \
        making the requests
        Args:
            datetime_obj: datetime.datetime object which needs to be converted

        Returns:
            binance specific format

        """
        return int(datetime_obj.timestamp()) * 1000


class SomeOtherTimeIntervalChunks(AbstractTimeIntervalChunks):
    @staticmethod
    def sanitize_datetime_to_exchange_specific(datetime_obj):
        pass



import logging
import asyncio
import pathlib
import os
from datetime import datetime
from typing import Dict, TypeVar
from abc import ABC
from dataclasses import dataclass


logger = logging.getLogger(__name__)


def init_logger(level=logging.INFO):
    log_format = "%(module)s : %(asctime)s : %(levelname)s : %(message)s"
    logging.basicConfig(level=level, format=log_format)


async def gather_with_concurrency(num_concurrent_tasks: int = None,
                                  *tasks):
    """
    Limit the number of tasks that are gather concurrently
    Args:
        num_concurrent_tasks(int): Number of concurrent tasks
        *tasks: Tasks to be gathered

    Returns: result of tasks

    """
    semaphore = asyncio.Semaphore(num_concurrent_tasks)

    async def sem_task(task):
        """
        If the num_concurrent_tasks is None, it does not \
         throttle and returns the task. If it is not None, it \
         throttles, the number of concurrent tasks
        Args:
            task: task which might have to be throttled

        Returns: result of task

        """
        if num_concurrent_tasks is None:
            return await task

        async with semaphore:
            return await task

    all_tasks = (sem_task(task) for task in tasks)
    return await asyncio.gather(*all_tasks)


async def gather_dict(tasks: dict):
    """
    Gathers the tasks in a dict format (instead of typical List)
    Args:
        tasks(dict): Dictionary of tasks to be gathered.
         key represents the identifier, the value is the task

    Returns: Result of the tasks in a dict format where key \
     is not modified from the tasks

    """
    async def mark(key, coroutine):
        return key, await coroutine

    tasks_to_send = (mark(key, coroutine)
                     for key, coroutine in tasks.items())

    return {
        key: result
        for key, result in await gather_with_concurrency(500,
                                                         *tasks_to_send
                                                         )
    }


class Borg:
    """Borg instance"""
    _shared_state = {}

    def __init__(self):
        self.__dict__ = self._shared_state


class TokenBucket(Borg):
    """Controls the number of requests that can be made to the API.
    All times are written in micro-seconds for a good level of accuracy"""

    def __init__(self,
                 request_limit: Dict,
                 pause_seconds: float = 1):
        """
        Initializes the TokenBucket algorithm to throttle the flow of requests

        Args:
            request_limit: dictionary of requests that can be made where \
            the key is the datetime.timedelta object and the value is the\
             number of maximum requests allowed
            pause_seconds: amount of seconds to pause and test again
        """
        super().__init__()
        if not self._shared_state:
            self.queue = asyncio.Queue()
            (
                self.bucket_list,
                self.delta_t_list,
                self.max_requests_list,
            ) = self._initialize_buckets(request_limit)
            self.last_check = datetime.now()
            self.pause_seconds = pause_seconds
            self.results = []
            self._counter = 0

    @staticmethod
    def _initialize_buckets(requests_dict: Dict):
        """
        Initializes the bucket with the tokens.
        Essentially fills up the buckets with various kinds of tokens

        # TODO Does not follow SRP. Split it into different methods
        Args:
            requests_dict(Dict):  the key is the datetime.timedelta object \
            and the value is the number of maximum requests allowed


        Returns:
            Tuple(List): Each item in the dict corresponds to each item\
             in the list
             bucket_list: which is the list of tokens inside the bucket.\
              During initialization it has the max number of tokens which\
               is equal to number of max-requests

            delta_t_list: which is the list of total-duration

            max_requests: which is the list of the max number of requests\
             in allotted time

        """
        bucket_list = []
        delta_t_list = []
        max_requests_list = []
        for delta_t, max_requests in requests_dict.items():
            bucket_list.append(max_requests)
            delta_t_list.append(delta_t.total_seconds() * 10e6)
            max_requests_list.append(max_requests)
        return bucket_list, delta_t_list, max_requests_list

    async def hold_if_exceeded(self):
        """Interface to the outside. It pauses the request \
        until the request is allowed"""
        await self._check_if_within_limits()
        self._counter += 1

    async def _check_if_within_limits(self):
        """Checks if the method can be called i.e if the requests is within the limits
        # TODO Does not follow SRP. Split it into parts"""
        current = datetime.now()
        time_passed = current - self.last_check
        self.last_check = current
        for it in range(len(self.bucket_list)):
            self.bucket_list[it] += (
                time_passed.total_seconds()
                * 10e6
                * self.max_requests_list[it]
                / self.delta_t_list[it]
            )

            if self.bucket_list[it] > self.max_requests_list[it]:
                self.bucket_list[it] = self.max_requests_list[it]
            if self.bucket_list[it] < 1:
                logger.debug(
                    "Requests have exceeded. Waiting for token"
                    " bucket to fill-up"
                )
                await asyncio.sleep(self.pause_seconds)
                if await self._check_if_within_limits() is True:
                    continue
            self.bucket_list[it] -= 1
        return True


class AbstractFactory(ABC):
    """Abstract Factory for all the factories which is responsible
     for registering classes"""

    _builders = {}

    @classmethod
    def register_builder(cls, factory_type, identifier, class_type):
        """Registers the factory to be used later by the user"""
        if factory_type not in cls._builders.keys():
            cls._builders[factory_type] = {}
        cls._builders[factory_type][identifier] = class_type

    @classmethod
    def get_builders(cls):
        return cls._builders


def register_factory(section, identifier):
    """Decorator for registering factories in the factory_types"""
    def decorate(decorated_class_type):
        """Registers the class type in the factory_type"""
        AbstractFactory.register_builder(section,
                                         identifier,
                                         decorated_class_type)

        class Wrapper:
            pass

        Wrapper.__doc__ = decorated_class_type.__doc__
        Wrapper.__name__ = decorated_class_type.__name__
        for attribute, func in decorated_class_type.__dict__.items():
            if callable(decorated_class_type.__dict__[attribute]):
                setattr(
                    Wrapper,
                    attribute,
                    decorated_class_type.__dict__[attribute],
                )
        # Necessary for sphinx to obtain the dc-string correctly
        return Wrapper

    return decorate


class RetryModel:
    """Provides the ability for the method to be retried"""

    def __init__(self, retries: int = 3, sleep_seconds: int = 5):
        """
        Initialize the retry model

        Args:
            retries(int): number of attempts
            sleep_seconds(int): number of seconds to sleep if the\
             retry failed
        """
        self.retries = retries
        self.sleep_seconds = sleep_seconds

    @property
    def retries(self):
        return self._retries

    @retries.setter
    def retries(self, value):
        if value < 1:
            logger.error("All retries have been consumed")
            raise ConnectionError("All retries have been consumed")
        self._retries = value

    async def consume_available_retry(self):
        """
        Triggered after an unsuccessful attempt.

        Returns:
            RetryModel: whose attempts to retry have been reduced by 1

        """
        await asyncio.sleep(self.sleep_seconds)
        return type(self)(retries=self._retries - 1)


def get_dataclass_from_dict(dataclass_name: str, dict_to_convert: Dict):
    """
    Converts a dict to a dataclass of required name
    Args:
        dataclass_name (str): name of the dataclass
        dict_to_convert (Dict): dictionary of items that need to be converted

    Returns:
        dataclass that is generated from the provided dictionary
    """
    dataclass_definition = type(
        dataclass_name,
        (),
        {"__annotations__": {k: type(v) for k, v in dict_to_convert.items()}},
    )
    dataclass_decorated = dataclass(dataclass_definition)
    return dataclass_decorated(**dict_to_convert)


TypeVarPlaceHolder = TypeVar("TypeVarPlaceHolder")


def check_for_write_access(path: pathlib.Path):
    return os.access(path, os.W_OK | os.X_OK)



import datetime
import re
import math
import pandas as pd


class DateTimeOperations:
    string_match_timedelta_dict = {
        "m": datetime.timedelta(minutes=1),
        "h": datetime.timedelta(hours=1),
        "d": datetime.timedelta(days=1),
        "w": datetime.timedelta(weeks=1),
        # 31 days chosen to not have insufficient lengths
        "M": datetime.timedelta(days=31),
    }

    def map_string_to_timedelta(self, time_string: str) -> datetime.timedelta:
        """
        Maps the string to timedelta
        Args:
            time_string: string which is supposed to represent time

        Returns:
            datetime.timedelta object of the string of time

        """
        if isinstance(time_string, float):
            if math.isnan(time_string):
                return pd.to_timedelta(time_string)
        number_of_items = int(re.search("[0-9]+", time_string).group())
        string_to_match = re.search("[a-zA-Z]+", time_string).group()
        try:
            timedelta_of_string = self.string_match_timedelta_dict[
                string_to_match
            ]
        except KeyError:
            raise KeyError(
                f"{string_to_match} could not match with anything "
                f"in the exchange."
            )
        return datetime.timedelta(
            seconds=(timedelta_of_string.total_seconds() *
                     number_of_items)
        )

    def map_string_to_seconds(self, time_string: str) -> float:
        """
        Maps the string to seconds
        Args:
            time_string: string which is supposed to represent time

        Returns:
            float: seconds of the time_string

        """
        return self.map_string_to_timedelta(time_string).total_seconds()



import asyncio
import logging
import time
from binance import exceptions
from binance.client import AsyncClient
from abc import ABC
from datetime import timedelta
from crypto_history.utilities.general_utilities import TokenBucket, RetryModel


logger = logging.getLogger(__name__)


class AbstractMarketRequester(ABC):
    """AbstractBaseClass for the low-level market requester"""

    def __init__(self):
        self._client = None
        self.retry_strategy_class = RetryModel
        self.request_queue = None

    async def __aenter__(self):
        return self

    async def __aexit__(self, *args):
        await self._client.close()

    async def request(self, method_name: str, *args, **kwargs):
        """
        Interface to the user of the low level request made to the API server

        Args:
            method_name: name of the method to be called on the client object.
            *args: arguments transferred by the market operator
            **kwargs: arguments transferred by the market operator

        Returns:
            response from the market/exchange

        """
        retry_strategy_state = self.retry_strategy_class()
        return await self._request_with_retries(
            method_name, retry_strategy_state, *args, **kwargs
        )

    @staticmethod
    def _log_request(method_name, *args, **kwargs):
        logger.debug(
            f"Requesting:\t"
            f"method: {method_name}\t"
            f"args: {args}\t"
            f"kwargs: {''.join(f'{key}: {value}' for key, value in kwargs)}"
        )

    async def _request_with_retries(
        self, method_name: str, retry_strategy_state=None, *args, **kwargs
    ):
        """
        This method is called when the intention is to give it the
         possibility to retry an operations if it times out

        Args:
            method_name: name of the method to be called on the client
            retry_strategy_state: RetryModel class instance
            *args: arguments for the request call
            **kwargs: keyword arguments for the request call

        Returns:
            response from the request

        """
        try:
            return await self._request(method_name, *args, **kwargs)
        except asyncio.exceptions.TimeoutError:
            return await self._retry(
                method_name, retry_strategy_state, *args, **kwargs
            )

    async def _retry(self, method_name, retry_strategy_state, *args, **kwargs):
        """
        Triggered when the original request had failed.

        Args:
            method_name: method for the request
            retry_strategy_state: current state of the RetryModel
            *args: arguments for the request
            **kwargs: keyword arguments for the request

        Returns:
            response from the request

        """
        logger.debug(
            f"Retrying {method_name} with args:{args}, kwargs:{kwargs}"
        )
        return await self._request_with_retries(
            method_name,
            await retry_strategy_state.consume_available_retry(),
            *args,
            **kwargs,
        )

    async def _request(self, method_name, *args, **kwargs):
        """
        The last layer of the request before making the call.
        Holds the request if the queue is close to exceeding the limit
        Args:
            method_name: method for the request
            *args: arguments for the request
            **kwargs: keyword arguments for the request

        Returns:

        """
        await self.request_queue.hold_if_exceeded()
        self._log_request(method_name, *args, **kwargs)
        return await getattr(self._client, method_name)(*args, **kwargs)


class BinanceRequester(AbstractMarketRequester):
    """Low level requester for the Binance API. api_key and \
    api_secret are not required to get market information.\
    Limit for requests are officially set at 2400 per minute. \
    However, it is throttled to 500 per minute
        """

    def __init__(self):
        super().__init__()
        self._client = AsyncClient(api_key="", api_secret="")
        self.request_queue = TokenBucket(
            request_limit={timedelta(minutes=1): 400}
        )

    async def _request_with_retries(
        self, method_name: str, retry_strategy_state=None, *args, **kwargs
    ):
        """
        Wrapper around the default _request_with_retries to avoid\
         failed requests after too many requests particular to Binance
        Args:
            method_name: name of the method to be called on the client
            retry_strategy_state: RetryModel class instance
            *args: arguments for the request call
            **kwargs: keyword arguments for the request call

        Returns:
            response from the request

        """
        try:
            return await super()._request_with_retries(
                method_name, retry_strategy_state, *args, **kwargs
            )
        except exceptions.BinanceAPIException as e:
            # Error code corresponds to TOO_MANY_REQUESTS in Binance
            if e.code == -1003:
                wait_seconds = 5
                logger.warning(
                    f"Request could not respond as TOO_MANY_REQUESTS. "
                    f"SYNCHRONOUSLY pausing everything for {wait_seconds} seconds. "
                    f"Reason {e}"
                )
                time.sleep(wait_seconds)
                logger.warning(
                    "Updating the Binance Session as it often gets stuck "
                    "without it"
                )

                return await self._retry(
                    method_name, retry_strategy_state, *args, **kwargs
                )
            raise


class SomeOtherExchangeRequester(AbstractMarketRequester):
    pass



from __future__ import annotations
import logging
from collections import UserList

logger = logging.getLogger(__name__)


class TickerPool(UserList):
    pass


class BinanceTickerPool(TickerPool):
    possible_reference_coins = {
        "bitcoin": ["BTC", "XBT"],
        "ethereum": ["ETH"],
        "binance": ["BNB"],
        "altcoins": ["XRP", "TRX"],
        "currency": [
            "USDT",
            "USDC",
            "BUSD",
            "TUSD",
            "EUR",
            "PAX",
            "USDS",
            "TRY",
            "RUB",
            "KRW",
            "IDRT",
            "GBP",
            "UAH",
            "IDR",
            "NGN",
            "ZAR",
        ],
    }

    futures_key_words = ("BEAR", "BULL", "UP", "DOWN")



# -*- coding: utf-8 -*-
from setuptools import setup

packages = \
['crypto_history',
 'crypto_history.data_container',
 'crypto_history.emit_data',
 'crypto_history.stock_market',
 'crypto_history.utilities']

package_data = \
{'': ['*']}

install_requires = \
['SQLAlchemy>=1.4.5,<2.0.0',
 'python-binance-async>=0.7.4-beta.0,<0.8.0',
 'xarray>=0.17.0,<0.18.0']

setup_kwargs = {
    'name': 'crypto-history',
    'version': '1.2b11',
    'description': 'crypto_history is a python package for extracting history of crypto-currencies from various exchanges and presenting them in a tabular data-format',
    'long_description': None,
    'author': 'Vikramaditya Gaonkar',
    'author_email': 'vikramaditya91@gmail.com',
    'maintainer': None,
    'maintainer_email': None,
    'url': None,
    'packages': packages,
    'package_data': package_data,
    'install_requires': install_requires,
    'python_requires': '>=3.8,<4.0',
}


setup(**setup_kwargs)



from __future__ import annotations
import logging
import contextlib
import xarray as xr
from pandas import DataFrame
from dataclasses import dataclass, fields
from typing import Union, List, Dict
from datetime import datetime
from crypto_history.stock_market.tickers import TickerPool
from crypto_history.stock_market.stock_market_factory import (
    StockMarketFactory,
    AbstractMarketHomogenizer
)
from crypto_history.data_container.utilities import DataFrameOperations
from crypto_history.utilities import general_utilities, exceptions

logger = logging.getLogger(__name__)


class PrimitiveCoinHistoryObtainer:
    """Abstract class to serve as the parent to generating \
    histories obtainer classes"""

    def __init__(
        self,
        market_harmonizer: AbstractMarketHomogenizer,
        interval: str,
        start_time: Union[str, datetime, int],
        end_time: Union[str, datetime, int],
    ):
        self.market_harmonizer = market_harmonizer
        self.data_container = None
        self.interval = interval
        self.start_time = start_time
        self.end_time = end_time
        self.ticker_pool = None
        self.example_raw_history = None

    @classmethod
    @contextlib.asynccontextmanager
    async def create_primitive_coin_history_obtainer(
        cls,
        exchange_factory: StockMarketFactory,
        interval: str,
        start_time: Union[str, datetime, int],
        end_time: Union[str, datetime, int],
    ) -> PrimitiveCoinHistoryObtainer:
        """
        Generates the coin history obtainer

        Args:
            exchange_factory(StockMarketFactory): instance of\
             the exchange_factory which is responsible for \
            setting the market_homogenizer
            interval(str): Length of the history of the klines per item
            start_time(str/datetime/int): duration from which history \
               is necessary
            end_time(str/datetime/int): duration up to which history \
               is necessary

        Yields:
            PrimitiveCoinHistoryObtainer instance from the above \
            arguments
        """
        async with exchange_factory.create_data_homogenizer() \
                as market_harmonizer:
            yield cls(
                market_harmonizer, interval, start_time, end_time,
            )

    async def initialize_example(self) -> List:
        # FixMe The TimeIndexedDataContainer something similar.\
        #  Should be centralized?
        """
        Initializes an example of the raw history and stores\
         it example_raw_history

        Returns:
             list: instance of an example of a raw history of ETHvsBTC

        """
        self.example_raw_history = self.example_raw_history or list(
            await self._get_raw_history_for_ticker("ETHBTC")
        )
        return self.example_raw_history

    async def _get_raw_history_for_ticker(self, ticker_symbol: str) -> map:
        """
        Gets the raw history for the ticker by using the interface\
         to the market provided by the market homogenizer

        Args:
            ticker_symbol(str): Symbol of the ticker whose history\
             is desired

        Returns:
             map: raw mapped history of the ticker mapped to the \
             OHLCVFields instance

        """
        return await self.market_harmonizer.get_history_for_ticker(
            ticker=ticker_symbol,
            interval=self.interval,
            start_time=self.start_time,
            end_time=self.end_time,
        )

    async def get_all_raw_tickers(self) -> TickerPool:
        """
        Obtains the ticker pool by accessing the market homogenizer

        Returns:
             TickerPool: The TickerPool filled with all the ticker\
              available to the market/exchange

        """
        return await self.market_harmonizer.get_all_raw_tickers()

    async def get_historical_data_from_base_and_reference_assets(
        self, base_assets: List, reference_assets: List
    ) -> Dict:
        """
        Get the historical data for the combination of base\
         and reference assets
        Args:
            base_assets (List['str']): list of all base assets
            reference_assets (List['str']): list of all reference assets

        Returns:
            The kline historical data of the combinations
        """
        raw_ticker_pool = await self.get_all_raw_tickers()
        tickers_to_capture = []
        for base_asset in base_assets:
            for reference_asset in reference_assets:
                if any(
                    f"{base_asset}{reference_asset}" == raw_ticker["symbol"]
                    for raw_ticker in raw_ticker_pool
                ):
                    tickers_to_capture.append((base_asset, reference_asset))
        return await self._get_historical_data_relevant_coins(
            tickers_to_capture
        )

    async def _get_historical_data_relevant_coins(
        self, tickers_to_capture: List[tuple]
    ) -> Dict:
        """
        Get the history of all the tickers in the market/exchange

        Args:
            tickers_to_capture (list): List of tickers to capture

        Returns:
             map: map of all the histories in the exchange/market
             mapped from history to the OHLCVFields

        """
        tasks_to_pursue = {}
        for base_asset, reference_asset in tickers_to_capture:
            tasks_to_pursue[
                base_asset, reference_asset
            ] = self._get_raw_history_for_ticker(
                f"{base_asset}{reference_asset}"
            )
        return await general_utilities.gather_dict(tasks_to_pursue)


class PrimitiveDimensionsManager:
    """Class responsible for managing the dimensions/coordinates \
    of the data container"""

    @dataclass
    class Dimensions:
        """Class for keeping track of the dimensions of the XDataArray"""

        base_assets: List
        reference_assets: List
        ohlcv_fields: List
        index_number: List

    def __init__(self, history_obtainer):
        self.coin_history_obtainer = history_obtainer

    @classmethod
    def create_primitive_dimensions_manager(
        cls, history_obtainer: PrimitiveCoinHistoryObtainer
    ) -> PrimitiveDimensionsManager:
        """Initializes the dimensions manager with the coin_history_obtainer"""
        return PrimitiveDimensionsManager(history_obtainer)

    async def get_depth_of_indices(self) -> List:
        """
        Obtains the depth of the indices to identify how many time-stamp values
         each history has available

        Returns:
             list: The list of indices from 0..n-1
                     where n corresponds to the number of time-stamps\
                      in the history

        """
        example_history = await self.coin_history_obtainer.initialize_example()
        indices = list(range(len(list(example_history))))
        return indices

    @staticmethod
    async def set_coords_dimensions_in_empty_container(
        dataclass_dimensions_coordinates,
    ) -> xr.DataArray:
        """
        Initialize the xarray container with None values but\
         with the coordinates as it helps in the memory allocation
        Args:
            dataclass_dimensions_coordinates: dataclass of \
            coordinates/dimensions as the framework for generating \
            the XDataArray
        """
        dimensions = [
            dimension.name
            for dimension in fields(dataclass_dimensions_coordinates)
        ]
        coordinates = [
            getattr(dataclass_dimensions_coordinates, item)
            for item in dimensions
        ]
        return xr.DataArray(None, coords=coordinates, dims=dimensions)

    async def get_dimension_coordinates_from_fields_and_assets(
        self, ohlcv_fields: List, base_assets: List, reference_assets: List
    ):
        """
        Gets the dataclass containing the coordinates required \
        for creating the xr.DataArray
        Args:
            ohlcv_fields: fields related to open_ts, close_ts, and so on
            base_assets: list of base coins
            reference_assets: list of reference coins

        Returns:
            DataClass of the coordinates necessary to build the DataArray
        """
        index_number = await self.get_depth_of_indices()
        return self.Dimensions(
            base_assets=base_assets,
            reference_assets=reference_assets,
            ohlcv_fields=ohlcv_fields + ["weight"],
            index_number=index_number,
        )


class PrimitiveDataArrayOperations:
    """Abstract Base Class for generating the data operations"""

    def __init__(
        self,
        history_obtainer: PrimitiveCoinHistoryObtainer,
        dimensions_manager: PrimitiveDimensionsManager,
        base_assets: List,
        reference_assets: List,
        ohlcv_fields: List,
        interval: str,
    ):
        self.history_obtainer = history_obtainer
        self.dimension_manager = dimensions_manager
        self.dataframe_operations = DataFrameOperations()
        self.base_assets = base_assets
        self.reference_assets = reference_assets
        self.ohlcv_fields = ohlcv_fields
        self.interval = interval

    @classmethod
    @contextlib.asynccontextmanager
    async def create_primitive_data_array_operations(
        cls,
        exchange_factory: StockMarketFactory,
        base_assets: List,
        reference_assets: List,
        ohlcv_fields: List,
        interval: str,
        start_time: Union[str, datetime, int],
        end_time: Union[str, datetime, int],
    ):
        """
        Initializes the DataContainerOperations which is the user \
        end-point for the basic data building
        Args:
            exchange_factory (StockMarketFactory): factory of the exchange
            base_assets (List): list of base coins to be accumulated
            reference_assets (List): list of reference/quote-against\
             coins to be accumulated
            ohlcv_fields (List): list of fields for the various fields
            interval (str): data capture interval
            start_time (str/datetime/int): date from which data collection \
                should start
            end_time (str/datetime/int): date up to which data collection \
                should be made
        Yields:
            PrimitiveDataArrayOperations instance from the above arguments
        """
        async with PrimitiveCoinHistoryObtainer.\
                create_primitive_coin_history_obtainer(
                exchange_factory, interval, start_time, end_time
                ) as history_obtainer:
            dimensions_manager = PrimitiveDimensionsManager.\
                create_primitive_dimensions_manager(
                    history_obtainer
                )
            yield cls(
                history_obtainer,
                dimensions_manager,
                base_assets,
                reference_assets,
                ohlcv_fields,
                interval,
            )

    async def get_populated_primitive_container(self) -> xr.DataArray:
        """
        Populates the container and returns it to the use

        Returns:
             xr.DataArray: the filled container of information

        """
        # TODO Avoid populating it every time it is called
        coord_dimension_dataclass = await self.dimension_manager.\
            get_dimension_coordinates_from_fields_and_assets(
                ohlcv_fields=self.ohlcv_fields,
                base_assets=self.base_assets,
                reference_assets=self.reference_assets,
            )

        data_container = await self.dimension_manager.\
            set_coords_dimensions_in_empty_container(
                coord_dimension_dataclass
            )
        data_container = await self.populate_container(
            data_container, coord_dimension_dataclass
        )
        return data_container

    @staticmethod
    def _insert_coin_history_in_container(
        data_container, base_asset, reference_asset, history_df
    ):
        """
        Low level function to inserts the coin history in the df
        Args:
            base_asset (str): base asset of the coin
            reference_asset (str): reference asset of the coin
            history_df (pandas.DataFrame): data frame of the coin history

        Returns:
            None
        """
        logger.debug(
            f"History set in x_array for ticker {base_asset}{reference_asset}"
        )
        data_container.loc[
            base_asset, reference_asset
        ] = history_df.transpose()
        return data_container

    @staticmethod
    def append_column_to_df(df, column_name, column_value) -> DataFrame:
        """
        Appends a new column to a df with constant value
        Args:
            df (DataFrame): dataframe to which column is to be appended
            column_name(str): name of the column
            column_value(str): value of the column

        Returns:
            new column added DataFrame
        """
        df[column_name] = column_value
        return df

    async def get_example_history(self) -> List:
        """A standard example on which the gaps are padded"""
        await self.history_obtainer.initialize_example()
        return self.history_obtainer.example_raw_history

    async def populate_container(
        self, data_container: xr.DataArray, coord_dimension_dataclass
    ) -> xr.DataArray:
        """
        Populates the xarray.DataArray with the historical \
        data of all the coins
        """
        historical_data = await self.history_obtainer.\
            get_historical_data_from_base_and_reference_assets(
                base_assets=coord_dimension_dataclass.base_assets,
                reference_assets=coord_dimension_dataclass.reference_assets,
            )
        standard_example_df = await self.get_example_history()
        for (
            (base_asset, reference_asset),
            ticker_raw_history,
        ) in historical_data.items():
            try:
                history_df = await self.dataframe_operations.get_compatible_df(
                    standard_example_df, ticker_raw_history
                )
            except exceptions.EmptyDataFrameException:
                logger.debug(
                    f"History does not exist for the combination "
                    f"of {base_asset}{reference_asset}"
                )
                continue
            else:
                history_df = self.dataframe_operations.\
                    drop_unnecessary_columns_from_df(
                        history_df, coord_dimension_dataclass.ohlcv_fields
                    )
                history_df = self.append_column_to_df(
                    history_df, "weight", self.interval
                )
            data_container = self._insert_coin_history_in_container(
                data_container, base_asset, reference_asset, history_df
            )
        return data_container



import logging
import itertools
import xarray as xr
import pandas as pd
import numpy as np
from typing import Union, Dict
from collections import OrderedDict
from crypto_history.stock_market.stock_market_factory import StockMarketFactory
from crypto_history.utilities import datetime_operations

logger = logging.getLogger(__package__)


class TypeConvertedData:
    """Type converts the data in the dataarray/dataset"""

    type_mapping = {
        int: pd.Int64Dtype,
        str: pd.StringDtype,
        float: np.float64,
    }

    def __init__(self,
                 exchange_factory: StockMarketFactory):
        """Initializes the type conversion of the data"""
        self.ohlcv_field_types = exchange_factory.create_ohlcv_field_types()

    def get_ohlcv_field_type_dict(self) -> Dict:
        """
        Gets the field types of the OHLCV Fields converted to \
        the numpy/pandas format. This is done to be able to \
        handle nan values in Int/String types

        Returns (dict): Dictionary of the map from the \
        ohlcv-field to the pd/np type

        """
        original_ohlcv_field_type = self.ohlcv_field_types.get_dict_name_type()
        original_ohlcv_field_type["weight"] = str
        return dict(
            map(
                lambda x: (x[0], self.type_mapping[x[1]]),
                original_ohlcv_field_type.items(),
            )
        )

    def set_type_on_dataset(self, dataset: xr.Dataset) -> xr.Dataset:
        """
        Sets the type on the xr.DataSet according to the \
        ohlcv field type

        Args:
            dataset(xr.DataSet): The dataset on which the type \
                has to be set

        Returns:
            xr.DataSet which has the type set on it
        """
        keys_in_dataset = list(dataset.keys())
        dict_of_types = self.get_ohlcv_field_type_dict()
        for key in keys_in_dataset:
            dataset[key] = dataset[key].astype(dict_of_types[key])
        return dataset

    def set_type_on_dataarray(self, dataarray: xr.DataArray) -> xr.DataArray:
        """
        Sets the type on the xr.DataArray according to the \
            ohlcv field type
        Args:
            dataarray(xr.DataArray): The DataArray on which the type \
                has to be set

        Returns:
            xr.DataArray which has the type set on it
        """
        dict_of_types = self.get_ohlcv_field_type_dict()
        for ohlcv_field in dataarray.ohlcv_fields.values:
            dataarray.loc[{"ohlcv_fields": ohlcv_field}] = dataarray.loc[
                {"ohlcv_fields": ohlcv_field}
            ].astype(dict_of_types[ohlcv_field])
        return dataarray


class HandleIncompleteData:
    """Responsible for handling missing data:\
    1. If a certain coin has to be dropped if it is null
    2. If a ticker has to be nulliifed as it has incomplete data
    """
    def __init__(self, coordinates_to_drop=None):
        """Initializes the incomplete data. The iterations on the
         coordinates are set with coordinates to drop"""
        if coordinates_to_drop is None:
            coordinates_to_drop = ["base_assets", "reference_assets"]
        self.coordinates_to_drop = coordinates_to_drop

    def drop_xarray_coins_with_entire_na(
        self, data_item: Union[xr.DataArray, xr.Dataset]
    ) -> Union[xr.DataArray, xr.Dataset]:
        """
        Drops the coins from the base/reference asset if all \
            its corresponding values are nan
        Args:
            data_item(xr.DataArray/xr.DataSet): which contains \
                information of the coin histories

        Returns:
            xr.DataArray/xr.DataSet where the coins have been dropped

        """
        logger.debug(f"{type(data_item)} prior to dropping entire "
                     f"data has coordinates: {data_item.coords}")
        for coordinate in self.coordinates_to_drop:
            data_item = data_item.dropna(coordinate, how="all")
        logger.debug(f"{type(data_item)} after dropping entire "
                     f"data has coordinates: {data_item.coords}")
        return data_item

    def get_all_coord_combinations(self,
                                   data_item: Union[xr.DataArray, xr.Dataset]):
        """
        Gets all the various combinations to iterate
         according to the coordinates to drop

        Args:
            data_item (xr.DataArray/xr.DataSet): data_item whose combinations
             need to be iterated over

        Yields:
            A dict with various combinations

        """
        combinations = OrderedDict()
        for coordinate in self.coordinates_to_drop:
            combinations[coordinate] = list(
                getattr(data_item, coordinate).values
            )
        keys, values = zip(*combinations.items())

        for combination in itertools.product(*values):
            yield dict(zip(keys, combination))

    def nullify_incomplete_data_from_dataarray(
        self, dataarray: xr.DataArray
    ) -> xr.DataArray:
        """
        Nullifies incomplete data from the xr.DataArray
        Args:
            dataarray (xr.DataArray): dataarray whose \
                coordinates are to be nullified

        Returns:
            xr.DataArray whose data has been nullified if incomplete

        """

        for combination in self.get_all_coord_combinations(dataarray):
            if dataarray.loc[combination].isnull().any():
                dataarray.loc[combination] = np.nan
        return dataarray

    def nullify_incomplete_data_from_dataset(
        self, dataset: xr.Dataset
    ) -> xr.Dataset:
        """
        Nullifies the incomplete data of datasets

        Notes:
            Using indexing to assign values to a
             subset of dataset (e.g., ds[dict(space=0)] = 1) is\
              not yet supported.
             http://xarray.pydata.org/en/stable/indexing.html
        Args:
            dataset(xr.DataSet): dataset whose data is to be nullified

        Returns:
            xr.DataSet whose incomplete items are nullified

        """
        for combination in self.get_all_coord_combinations(dataset):
            if any(list(dataset.loc[combination].isnull().any().values())):
                for ohlcv_field in list(dataset.keys()):
                    dataset[ohlcv_field].loc[combination] = np.nan
        return dataset


class ApplyWeightToDataArray:
    @staticmethod
    def get_weights_numpy_float_array(numpy_array: np.ndarray) -> np.ndarray:
        """
        Obtains the numpy array in float values from the string
        Args:
            numpy_array: np.ndarray numpy array of strings

        Returns: np.ndarray numpy array of seconds (float values)

        """
        date_time_operator = datetime_operations.DateTimeOperations()
        string_to_seconds = date_time_operator.map_string_to_seconds
        return np.array(
            list(
                map(lambda x: np.vectorize(string_to_seconds)(x), numpy_array))
        )

    @staticmethod
    def correct_shaped_np_array(numpy_array: np.ndarray,
                                len_ohlcv_fields: int) -> np.ndarray:
        """
        Fix the shape of the np array by filling in the shape
        Args:
            numpy_array: np.ndarray whose shape is not ok
            len_ohlcv_fields: int of the duplication required

        Returns: np.ndarray whose shape is as expected for the da

        """
        # TODO This should ideally do setting indices programatically
        numpy_array_dim_added = np.expand_dims(numpy_array, 3)
        return np.tile(numpy_array_dim_added, (1, 1, 1, len_ohlcv_fields))

    def get_weights_dataarray(self,
                              original_dataarray: xr.DataArray
                              ) -> xr.DataArray:
        """
        Gets the da of weights
        Args:
            original_dataarray: xr.DataArray which is not weighted,\
             but has a "weight" data_var

        Returns: xr.DataArray which is weighted

        """
        numpy_weights = self.get_weights_numpy_float_array(
            original_dataarray.loc[:, :, :, "weight"]
        )
        well_shaped_np_array = self.correct_shaped_np_array(
            numpy_weights,
            len(original_dataarray.ohlcv_fields)
        )
        return xr.DataArray(
            well_shaped_np_array,
            dims=original_dataarray.dims,
            coords=original_dataarray.coords
        )

    def get_weighted_data_container(self,
                                    original_dataarray: xr.DataArray):
        """
        Gets the weighted xr.DataArray
        Args:
            original_dataarray: xr.DataArray which has to be weighted

        Returns: xr.DataArray which is weighted

        """
        weights_dataarray = self.get_weights_dataarray(original_dataarray)
        return original_dataarray.weighted(weights_dataarray)

    def switch_str_to_float(self,
                            original_da: xr.DataArray
                            ) -> xr.DataArray:
        """
        Switches a dataarray of strings to dataarray of floats
        Args:
            original_da: xr.DataArray whose values are str

        Returns: xr.DataArray whose values are float

        """
        float_np = self.get_weights_numpy_float_array(original_da.values)
        return xr.DataArray(
            float_np,
            dims=original_da.dims,
            coords=original_da.coords
        )



from .utilities.general_utilities import init_logger, AbstractFactory  # NOQA
from .data_container import data_container_access  # NOQA
from .stock_market import stock_market_factory  # NOQA
from .data_container import data_container_post # NOQA
from .emit_data import save_to_disk # NOQA

class_builders = AbstractFactory.get_builders()



import pandas as pd
from typing import List, Iterable
from pandas import DataFrame

from crypto_history.utilities import exceptions


class DataFrameOperations:
    """Operations purely dedicated to dataframe"""

    @staticmethod
    def calculate_rows_to_add(
        df: DataFrame, list_of_standard_history: List
    ) -> int:
        """
        Calculates the additional number of rows that might have \
        to be added to get the df in the same shape

        Args:
            df(pd.DataFrame): pandas dataframe which is obtained\
             for the coin's history
            list_of_standard_history: expected standard history\
             which has the complete history

        Returns:
             int: number of rows that have to be added to the df

        """
        df_rows, _ = df.shape
        expected_rows = len(list_of_standard_history)
        return expected_rows - df_rows

    @staticmethod
    def drop_unnecessary_columns_from_df(
        df, necessary_columns: List
    ) -> DataFrame:
        """
        Drop all columns which are not necessary from the df
        Args:
            df (pd.DataFrame): from which the unnecessary columns\
             are to be dropped
            necessary_columns (list): list of columns which are\
             to be stored

        Returns:
            pd.DataFrame where the unnecessary columns are dropped
        """
        unnecessary_columns = [
            col for col in df.columns if col not in necessary_columns
        ]
        return df.drop(unnecessary_columns, axis=1)

    async def get_compatible_df(
        self, standard_example: List, ticker_history: Iterable
    ) -> DataFrame:
        """
        Makes the ticker history compatible to the standard \
        pd.DataFrame by extending the shape to add null values

        Args:
            standard_example (DataFrame): standard example to know \
            how many rows to pad
            ticker_history: history of the current ticker

        Returns:
             pd.DataFrame: compatible history of the df adjusted for \
             same rows as expected

        """
        # TODO Assuming that the df is only not filled in the bottom
        history_df = pd.DataFrame(ticker_history)
        if history_df.empty:
            raise exceptions.EmptyDataFrameException
        padded_df = await self.pad_extra_rows_if_necessary(
            standard_example, history_df
        )
        return padded_df

    @staticmethod
    def add_extra_rows_to_bottom(df: pd.DataFrame, empty_rows_to_add: int):
        """
        Adds extra rows to the pd.DataFrame

        Args:
            df(pd.DataFrame): to which extra rows are to be added
            empty_rows_to_add(int): Number of extra rows that have to be added

        Returns:
             pd.DataFrame: Re-indexed pd.DataFrame which has \
             the compatible number of rows

        """
        new_indices_to_add = list(
            range(df.index[-1] + 1, df.index[-1] + 1 + empty_rows_to_add)
        )
        return df.reindex(df.index.to_list() + new_indices_to_add)

    async def pad_extra_rows_if_necessary(
        self, standard_example: List, history_df: DataFrame
    ) -> DataFrame:
        """
        Add extra rows on the bottom of the DF is necessary. i.e \
        when the history is incomplete.
        # FixMe . Probably needs to be reevaluated
        Args:
            standard_example (DataFrame): standard example on which\
             it is based
            history_df: history of the dataframe that has not been\
             padded yet

        Returns:
            pd.DataFrame: that has been padded

        """
        rows_to_add = self.calculate_rows_to_add(history_df, standard_example)
        if rows_to_add > 0:
            history_df = self.add_extra_rows_to_bottom(history_df, rows_to_add)
        return history_df
